from trallie.providers import get_provider
from trallie.providers import ProviderInitializationError
from trallie.prompts import (
    ZERO_SHOT_EXTRACTION_SYSTEM_PROMPT,
    FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_DE,
    FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_FR,
    FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_ES,
    FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_IT
)
from trallie.data_handlers import DataHandler

import json

import re

# Post processing for a reasoning model 
def post_process_response(response: str) -> str:
    """
    Removes <think>...</think> content from the response.
    """
    return re.sub(r"<think>.*?</think>", "", response, flags=re.DOTALL).strip()

class DataExtractor:
    LANGUAGE_PROMPT_MAP = {
        "en": ZERO_SHOT_EXTRACTION_SYSTEM_PROMPT,
        "de": FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_DE,
        "fr": FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_FR,
        "es": FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_ES,
        "it": FEW_SHOT_EXTRACTION_SYSTEM_PROMPT_IT,
    }

    ALLOWED_NON_EN_MODELS = {"gpt-4o", "llama-3.3-70b-versatile"}
    ALLOWED_NON_EN_PROVIDERS = {"openai", "groq"}
    ALLOWED_REASONING_MODELS = {"deepseek-r1-distill-llama-70b"}
    
    def __init__(self, provider, model_name, system_prompt=None, language="en", reasoning_mode=False):
        self.provider = provider
        self.model_name = model_name
        self.client = get_provider(self.provider)
        self.language = language
        self.reasoning_mode = reasoning_mode

        if self.reasoning_mode and self.model_name not in self.ALLOWED_REASONING_MODELS:
            raise ValueError(
                f"`reasoning_mode=True` is not supported for model '{self.model_name}'. "
            )
    
        if self.language == "en":
            self.system_prompt = system_prompt or self.LANGUAGE_PROMPT_MAP["en"]
        else:
            # Enforce allowed providers/models for non-English
            if self.provider not in self.ALLOWED_NON_EN_PROVIDERS:
                raise ValueError(f"Provider '{self.provider}' is not supported for language '{self.language}'.")

            if self.model_name not in self.ALLOWED_NON_EN_MODELS:
                raise ValueError(f"Model '{self.model_name}' is not allowed for non-English extraction.")

            self.system_prompt = system_prompt or self.LANGUAGE_PROMPT_MAP.get(self.language)
            if not self.system_prompt:
                raise ValueError(f"No prompt available for language '{self.language}'.")

    def extract_attributes(self, schema, record, max_retries=3):
        """
        Extracts attributes for a given record and schema.
        """
        user_prompt = f"""
            Following is the record: {record} and the attribute schema for extraction: {schema}
            Provide the extracted attributes. Avoid any words at the beginning and end.
        """
        for attempt in range(max_retries):
            try:
                response = self.client.do_chat_completion(
                    self.system_prompt, user_prompt, self.model_name
                )
                # Validate if response is a valid JSON
                # print(response)
                if self.reasoning_mode:
                    response = post_process_response(response)
                response = json.loads(response)
                return response
            except (json.JSONDecodeError, TypeError) as e:
                print(f"Invalid JSON response (attempt {attempt + 1}): {e}")
                if attempt == max_retries - 1:
                    return None
            except Exception as e:
                print(f"Error: {e}")
                return None

    def extract_data(self, schema, record, max_retries=3, from_text=False):
        """
        Processes record and returns extracted attributes.
        """
        record_text = DataHandler(record, from_text=from_text).get_text()
        return self.extract_attributes(schema, record_text, max_retries)



import pandas as pd
import json
from typing import List, Tuple
from datasets import Dataset
import difflib
from collections import defaultdict


################### SCHEMA GENERATION HELPERS ###########################
def group_descriptions(descriptions, group_size):
    """
    Groups descriptions into batches of a given size.
    """
    return [
        descriptions[i : i + group_size]
        for i in range(0, len(descriptions), group_size)
    ]


def validate_json(output):
    """
    Validates if the given string is a valid JSON.
    """
    try:
        json.loads(output)
        return True
    except json.JSONDecodeError:
        return False


def extract_keys(schema, parent_key=None):
    """
    Recursively extract all keys from a JSON schema.
    """
    keys = set()
    for key, value in schema.items():
        full_key = f"{parent_key}.{key.lower()}" if parent_key else key.lower()
        keys.add(full_key)
        if isinstance(value, dict):
            keys.update(extract_keys(value, full_key))
    return keys


def prepare_eval_input(data_collection, sentences_per_group=20):
    eval_input = []
    current_group = []
    record_counter = 1  # Start record count at 1

    for example in data_collection["validation"]:
        tokens = example["tokens"]
        # Create the sentence with the "Record x:" format
        sentence = f"Record {record_counter}: " + " ".join(tokens)
        current_group.append(sentence)
        record_counter += 1  # Increment record counter for each sentence

        # Check if we reached the desired number of sentences in the group
        if len(current_group) >= sentences_per_group:
            # Join sentences in the group with a newline
            eval_input.append("\n".join(current_group))
            current_group = []  # Reset for the next group

    # Add any remaining sentences as the last group if not empty
    if current_group:
        eval_input.append("\n".join(current_group))

    return eval_input


def filter_duplicate_text_columns(dataset, column_name):
    """
    Filters duplicate text columns in a dataset.

    Args:
      dataset: The input dataset (Hugging Face Datasets format).
      column_name: The name of the text column to filter.

    Returns:
      A new dataset with duplicate text entries removed from the specified column.
    """

    # Convert the dataset column to a list
    text_column = dataset[column_name]

    # Efficiently find unique entries and their indices
    unique_texts = []
    unique_indices = []
    seen_texts = set()
    for index, text in enumerate(text_column):
        if text not in seen_texts:
            unique_texts.append(text)
            unique_indices.append(index)
            seen_texts.add(text)

    # Create a new dataset with only the unique entries
    filtered_dataset = Dataset.from_dict({column_name: unique_texts})

    return filtered_dataset


def remove_duplicate_chunks(text):
    """Remove duplicate chunks of text while preserving order."""
    words = text.split()
    seen = set()
    filtered_words = []

    for word in words:
        if word not in seen:
            seen.add(word)
            filtered_words.append(word)

    return " ".join(filtered_words)


def group_text_by_file_name(dataset):
    grouped = defaultdict(list)

    for row in dataset["validaion"]:  # Fixed typo
        grouped[row["file_name"]].append(row["text"])  # Group texts by file name

    # Remove duplicate phrases and return a list of deduplicated full texts
    return [remove_duplicate_chunks(" ".join(texts)) for texts in grouped.values()]


########### SCHEMA FILLING HELPERS #####################
def group_text_and_values_by_file_name(dataset):
    grouped = {}

    for row in dataset["validaion"]:  # Fixed typo
        file_name = row["file_name"]
        if file_name not in grouped:
            grouped[file_name] = {"keys": [], "values": [], "full_text": []}

        grouped[file_name]["keys"].append(row["key"])
        grouped[file_name]["values"].append(row["value"])
        grouped[file_name]["full_text"].append(row["text"])  # Collect all text chunks

    # Remove duplicate phrases and return cleaned data
    return [
        {
            "file_name": key,
            **value,
            "full_text": remove_duplicate_chunks(
                " ".join(value["full_text"])
            ),  # Clean duplicate chunks
        }
        for key, value in grouped.items()
    ]



import json
from bs4 import BeautifulSoup
from PyPDF2 import PdfReader

def infer_datatype(func):
    """
    Decorator to infer the datatype of the document and set it.
    """

    def wrapper(self, *args, **kwargs):
        self.datatype = self.document.split(".")[-1].lower()
        return func(self, *args, **kwargs)

    return wrapper
    
class DataHandler:
    def __init__(self, document, from_text=False):
        self.document = document
        self.from_text = from_text
        self.datatype = None
        self.length = None  # Length will be set by the decorator
        self.text = None  # Contains extracted text

    def get_text_from_pdf(self):
        # Use a pdf extractor
        try:
            # Open and read the PDF file
            reader = PdfReader(self.document)
            text = ""
            # Iterate through all pages and extract text
            for page in reader.pages:
                text += page.extract_text()
            return text.strip() if text else "Error: No text found in the PDF"
        except FileNotFoundError:
            return "Error: File not found"
        except Exception as e:
            return f"Error: An unexpected error occurred: {str(e)}"
        return "Extracted text from PDF"

    def get_text_from_html(self):
        # Use an HTML parser
        try:
            with open(self.document, "r", encoding="utf-8") as file:
                return file.read()
        except FileNotFoundError:
            return "Error: File not found"
        except Exception as e:
            return f"Error: An unexpected error occurred: {str(e)}"
        return "Extracted text from HTML"

    def get_text_from_json(self):
        # Extract text from JSON
        try:
            with open(self.document, "r", encoding="utf-8") as file:
                data = json.load(file)
            # Convert the JSON object to a pretty-printed string for readability
            return json.dumps(data, indent=4)
        except FileNotFoundError:
            return "Error: File not found"
        except json.JSONDecodeError:
            return "Error: Invalid JSON format"
        return "Extracted text from JSON"

    def get_text_from_txt(self):
        try:
            with open(self.document, "r") as file:
                return file.read()
        except FileNotFoundError:
            return "Error: File not found"
        except Exception as e:
            return f"Error: An unexpected error occurred: {str(e)}"

    def chunk_text(self, char_limit=100000):
        """
        Decorator to chunk longer document if it exceeds the word length.
        """
        if self.length > char_limit:
            self.text = self.text[:char_limit]
        return self.text

    @infer_datatype
    def get_text(self):
        """
        Get text from the document based on its datatype.
        """
        if self.from_text:
            self.datatype = "txt"
            self.text = self.document
        else:
            if self.datatype == "pdf":
                self.text = self.get_text_from_pdf()
            elif self.datatype in {"html", "htm"}:
                self.text = self.get_text_from_html()
            elif self.datatype == "json":
                self.text = self.get_text_from_json()
            elif self.datatype == "txt":
                self.text = self.get_text_from_txt()
            else:
                return "Unsupported file type"

        # Chunk text
        self.length = len(self.text)
        self.chunk_text()
        return self.text



import json
import os
from collections import defaultdict
from tqdm import tqdm
from sentence_transformers import SentenceTransformer, util

def embedding_match(text1, text2, model, threshold=0.5, use_exact_match=False):
    # Return 0 if either text is missing
    if not text1 or not text2:
        return 0.0, False

    # If exact match is enabled, check first
    if use_exact_match and text1.strip().lower() == text2.strip().lower():
        return 1.0, True

    # Fall back to S-BERT similarity
    embeddings = model.encode([text1, text2], convert_to_tensor=True)
    score = util.cos_sim(embeddings[0], embeddings[1]).item()
    return score, score >= threshold

# def embedding_sim_sbert(text1, text2, model, threshold=0.5):
#     # Compute S-BERT between two pieces of text
#     if not text1 or not text2:
#         return 0.0, False
#     embeddings = model.encode([text1, text2], convert_to_tensor=True)
#     score = util.cos_sim(embeddings[0], embeddings[1]).item()
#     return score, score >= threshold

def evaluate_closedie(ground_truth_path, predicted_path, value_threshold=0.5, use_exact_match=False):
    # Evaluate performance of ClosedIE task
    model = SentenceTransformer("all-MiniLM-L6-v2")
    with open(ground_truth_path, "r") as f:
        ground_truth = json.load(f)
    with open(predicted_path, "r") as f:
        predicted = json.load(f)
    predicted = {
        k: json.loads(v) if isinstance(v, str) else v for k, v in predicted.items()
    }
    ground_truth = {os.path.basename(k): v for k, v in ground_truth.items()}
    common_files = set(ground_truth.keys()) & set(predicted.keys())
    entity_metrics = defaultdict(lambda: {"tp": 0, "fp": 0, "fn": 0})
    for filename in tqdm(common_files):
        gt_entities = ground_truth[filename] or {}
        pred_entities = predicted.get(filename) or {}
        for gt_key, gt_vals in gt_entities.items():
            gt_values = gt_vals if isinstance(gt_vals, list) else [gt_vals]
            pred_raw = pred_entities.get(gt_key, [])
            pred_values = pred_raw if isinstance(pred_raw, list) else [pred_raw]
            if not pred_values:
                entity_metrics[gt_key]["fn"] += len(gt_values)
                continue
            matched_preds = set()
            for gt_value in gt_values:
                match_found = False
                for idx, pred_value in enumerate(pred_values):
                    if idx in matched_preds:
                        continue
                    score, match = embedding_match(
                        gt_value, pred_value, model, value_threshold, use_exact_match=use_exact_match
                    )
                    if match:
                        entity_metrics[gt_key]["tp"] += 1
                        matched_preds.add(idx)
                        match_found = True
                        break
                if not match_found:
                    entity_metrics[gt_key]["fn"] += 1
            entity_metrics[gt_key]["fp"] += len(pred_values) - len(matched_preds)
    return _compute_f1_metrics(entity_metrics)

def find_best_matching_key(gt_key, pred_keys, model, key_threshold=0.8):
    if not pred_keys:
        return None, 0.0
    gt_embedding = model.encode(gt_key, convert_to_tensor=True)
    pred_embeddings = model.encode(pred_keys, convert_to_tensor=True)
    scores = util.cos_sim(gt_embedding, pred_embeddings)[0].cpu().numpy()
    best_idx = scores.argmax()
    best_score = scores[best_idx]
    return (
        (pred_keys[best_idx], best_score)
        if best_score >= key_threshold
        else (None, 0.0)
    )

def evaluate_openie(ground_truth_path, predicted_path, key_threshold=0.5, value_threshold=0.5, use_exact_match=False):
    # Evaluate OpenIE task
    model = SentenceTransformer("all-MiniLM-L6-v2")
    with open(ground_truth_path, "r") as f:
        ground_truth = json.load(f)
    with open(predicted_path, "r") as f:
        predicted = json.load(f)
    predicted = {
        k: json.loads(v) if isinstance(v, str) else v for k, v in predicted.items()
    }
    ground_truth = {os.path.basename(k): v for k, v in ground_truth.items()}
    common_files = set(ground_truth.keys()) & set(predicted.keys())
    entity_metrics = defaultdict(lambda: {"tp": 0, "fp": 0, "fn": 0})
    for filename in tqdm(common_files):
        gt_entities = ground_truth[filename] or {}
        pred_entities = predicted.get(filename) or {}
        pred_keys = list(pred_entities.keys())
        for gt_key, gt_value in gt_entities.items():
            best_pred_key, key_score = find_best_matching_key(
                gt_key, pred_keys, model, key_threshold
            )
            if best_pred_key:
                pred_value = pred_entities[best_pred_key]
                value_score, value_match = embedding_match(
                    gt_value, pred_value, model, value_threshold, use_exact_match=use_exact_match
                )
                if value_match:
                    entity_metrics[gt_key]["tp"] += 1
                else:
                    entity_metrics[gt_key]["fp"] += 1
                    entity_metrics[gt_key]["fn"] += 1
            else:
                entity_metrics[gt_key]["fn"] += 1
    return _compute_f1_metrics(entity_metrics)

def _compute_f1_metrics(entity_metrics):
    entity_f1_scores = {}
    all_tps, all_fps, all_fns = 0, 0, 0
    for entity, metrics in entity_metrics.items():
        tp, fp, fn = metrics["tp"], metrics["fp"], metrics["fn"]
        precision = tp / (tp + fp) if (tp + fp) > 0 else 0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0
        f1 = (2 * precision * recall / (precision + recall)) if (precision + recall) > 0 else 0
        entity_f1_scores[entity] = f1
        all_tps += tp
        all_fps += fp
        all_fns += fn
    macro_f1 = (
        sum(entity_f1_scores.values()) / len(entity_f1_scores)
        if entity_f1_scores
        else 0
    )
    micro_precision = all_tps / (all_tps + all_fps) if (all_tps + all_fps) > 0 else 0
    micro_recall = all_tps / (all_tps + all_fns) if (all_tps + all_fns) > 0 else 0
    micro_f1 = (
        (2 * micro_precision * micro_recall / (micro_precision + micro_recall))
        if (micro_precision + micro_recall) > 0
        else 0
    )
    return {
        "entity_f1_scores": entity_f1_scores,
        "macro_f1": macro_f1,
        "micro_f1": micro_f1,
        "micro_precision": micro_precision,
        "micro_recall": micro_recall,
    }



from trallie.providers import get_provider
from trallie.providers.groq import GroqProvider
from trallie.providers.openai import OpenAIProvider
from trallie.providers.ollama import OllamaProvider
from trallie.providers.huggingface import HuggingFaceProvider



import os

from trallie.schema_generation.schema_generator import SchemaGenerator
from trallie.data_extraction.data_extractor import DataExtractor

os.environ["GROQ_API_KEY"] = None #ENTER GROQ KEY HERE
os.environ["OPENAI_API_KEY"] = None #ENTER OPENAI KEY HERE

# Define the path to a set of documents/a data collection for inference
records = [
    "data/use-cases/EO_papers/pdf_0808.3837.pdf",
    "data/use-cases/EO_papers/pdf_1001.4405.pdf",
    "data/use-cases/EO_papers/pdf_1002.3408.pdf",
]

# Provide a description of the data collection
description = "A dataset of Earth observation papers"

# Initialize the schema generator with a provider and model
schema_generator = SchemaGenerator(provider="openai", model_name="gpt-4o")
# Feed records to the LLM and discover schema
print("SCHEMA GENERATION IN ACTION ...")
schema = schema_generator.discover_schema(description, records)
print("Inferred schema", schema)

# Initialize data extractor with a provider and model
data_extractor = DataExtractor(provider="openai", model_name="gpt-4o")
# Extract values from the text based on the schema
print("SCHEMA COMPLETION IN ACTION ...")
for record in records:
    extracted_json = data_extractor.extract_data(schema, record)
    print("Extracted attributes:", extracted_json)



# InLegalNER gold schema
inlegalner_gold_entity_schema = {
    "lawyer": "lawyer mentioned in the document",
    "court": "court involved in the case",
    "judge": "judge presiding over the case",
    "petitioner": "person or entity filing the petition",
    "respondent": "person or entity responding to the petition",
    "case_number": "unique identifier for the case",
    "gpe": "geopolitical entity mentioned in the judgment",
    "date": "date of the judgment or case events",
    "org": "organization mentioned in the judgment",
    "statute": "legal statute referred to in the judgment",
    "witness": "person testifying in the case",
    "precedent": "previous case law referenced in the judgment",
    "provision": "specific provision of law cited",
    "other_person": "any other person mentioned in the judgment",
}

# MIT Restaurants NER gold data
restaurant_data_gold_schema = {
    "rating": "customer rating of the restaurant",
    "amenity": "facilities or services provided by the restaurant",
    "location": "geographical location of the restaurant",
    "restaurant_name": "name of the restaurant",
    "price": "cost range of the restaurant",
    "hours": "operating hours of the restaurant",
    "dish": "specific dish served by the restaurant",
    "cuisine": "type of cuisine offered by the restaurant",
}

# MIT Movies NER gold schema
movie_data_gold_schema = {
    "actor": "actor appearing in the movie",
    "plot": "summary or storyline of the movie",
    "opinion": "review or opinion about the movie",
    "award": "award won or nominated for the movie",
    "year": "release year of the movie",
    "genre": "category or genre of the movie",
    "origin": "country or region of the movie's production",
    "director": "director of the movie",
    "soundtrack": "music or soundtrack featured in the movie",
    "relationship": "character relationships in the movie",
    "character_name": "name of a character in the movie",
    "quote": "memorable quote from the movie",
}

# WNUT-17 NER gold schema
wnut17_entity_gold_schema = {
    "person": "an individual mentioned in the context",
    "location": "a geographical place or address",
    "group": "a collection of individuals or an organization",
    "corporation": "a legally recognized business entity",
    "product": "a manufactured or digital item",
    "creative-work": "a work of art, literature, film, or other creative content",
}

# CADEC Schema
medical_entity_schema = {
    "Adverse drug reaction": "adverse drug reaction caused by medication",
    "Disease": "a medical condition affecting health",
    "Drug": "a pharmaceutical substance used for treatment",
    "Finding": "a clinical or diagnostic observation",
    "Symptom": "a physical or mental indication of a condition",
}

# Industry data gold schema
industry_data_entity_gold_schema = {
    "location": "geographical location of the project",
    "industry": "sector or type of industry",
    "productionstatus": "current production status",
    "assettype": "type of asset",
    "productionvolume": "current production volume",
    "revenuegoal": "expected revenue",
    "investmentask": "amount of investment required",
    "propertyvalue": "value of the property or asset",
    "reserves": "amount of proven reserves",
}

# FDA Data Schema
medical_device_submission_schema = {
    "product code": "unique identifier for the medical device category",
    "purpose for submission": "reason for submitting the application",
    "predicate device name": "name of the previously approved device used for comparison",
    "type of test": "category of testing performed on the device",
    "measurand": "specific quantity or parameter measured by the device",
    "proprietary and established names": "brand and official names of the device",
    "proposed labeling": "suggested labeling and instructions for the device",
    "conclusion": "final assessment or decision regarding the device submission",
    "indications for use": "specific medical conditions the device is intended to address",
    "classification": "regulatory classification of the device",
    "applicant": "entity or individual submitting the application",
    "panel": "FDA advisory panel reviewing the submission",
    "intended use": "primary purpose and function of the device",
    "regulation section": "specific regulatory code governing the device",
    "510k number": "unique identifier for the 510(k) submission",
}



import os
import glob

from trallie.evaluation.evaluation import evaluate_openie, evaluate_closedie
from trallie.evaluation.evaluation_params import get_evaluation_params, get_dataset_schema
from trallie.wrappers import openie, closedie

from pathlib import Path

# os.environ["GROQ_API_KEY"] = None
# os.environ["OPENAI_API_KEY"] = None 

os.environ["GROQ_API_KEY"] = "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"
os.environ["OPENAI_API_KEY"] = "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"

base_path = Path("data/evaluation/evaporate/data")
datasets = get_evaluation_params(base_path)

model_names = ["llama-3.3-70b-versatile", "gpt-4o", "llama-3.1-8b-instant", "deepseek-r1-distill-llama-70b"]
model_providers = ["groq", "openai", "groq", "groq"]
reasoning_modes = [False, False, False, True]

for dataset_name, dataset_info in list(datasets.items())[:1]:
    for provider, model_name, reasoning_mode in zip(model_providers, model_names, reasoning_modes):
        dataset_info["schema"] = get_dataset_schema(dataset_info["ground_truth"])
        records = glob.glob(str(dataset_info["files"]))
        if not os.path.exists(f"results/{model_name}_{dataset_name}_openie_predicted_table.json"):
            extracted_jsons = openie(dataset_info['description'], records, provider=provider, model_name=model_name, reasoning_mode=reasoning_mode, dataset_name=dataset_name)
        result = evaluate_openie(dataset_info['ground_truth'], f"results/{model_name}_{dataset_name}_openie_predicted_table.json", use_exact_match=True)
        print(f"Results for OpenIE {model_name} : {result}")
        if not os.path.exists(f"results/{model_name}_{dataset_name}_closedie_predicted_table.json"):
            schema = get_dataset_schema(dataset_info['ground_truth'])
            extracted_jsons = closedie(records, schema, provider=provider, model_name=model_name, reasoning_mode=reasoning_mode, dataset_name=dataset_name)
        result = evaluate_closedie(dataset_info["ground_truth"], f"results/{model_name}_{dataset_name}_closedie_predicted_table.json", use_exact_match=True)
        print(f"Results for ClosedIE {model_name} : {result}")
    